#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# melusine documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

import melusine
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
# import sphinx_bootstrap_theme

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.viewcode', 'numpydoc']
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Melusine'
copyright = u"Maif"
author = u"Sacha Samama & Tom Stringer"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = melusine.__version__
# The full version, including alpha/beta/rc tags.
release = melusine.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
# html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
# theme further.
# html_theme_options = {
#     # Navigation bar title. (Default: ``project`` value)
#     'navbar_title': "Demo",
#
#     # Tab name for entire site. (Default: "Site")
#     'navbar_site_name': "Site",
#
#     # A list of tuples containing pages or urls to link to.
#     # Valid tuples should be in the following forms:
#     #    (name, page)                 # a link to a page
#     #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
#     #    (name, "http://example.com", True) # arbitrary absolute url
#     # Note the "1" or "True" value above as the third argument to indicate
#     # an arbitrary url.
#     'navbar_links': [
#         ("Examples", "examples"),
#         ("Link", "http://example.com", True),
#     ],
#
#     # Render the next and previous page links in navbar. (Default: true)
#     'navbar_sidebarrel': True,
#
#     # Render the current pages TOC in the navbar. (Default: true)
#     'navbar_pagenav': True,
#
#     # Tab name for the current pages TOC. (Default: "Page")
#     'navbar_pagenav_name': "Page",
#
#     # Global TOC depth for "site" navbar tab. (Default: 1)
#     # Switching to -1 shows all levels.
#     'globaltoc_depth': 2,
#
#     # Include hidden TOCs in Site navbar?
#     #
#     # Note: If this is "false", you cannot have mixed ``:hidden:`` and
#     # non-hidden ``toctree`` directives in the same page, or else the build
#     # will break.
#     #
#     # Values: "true" (default) or "false"
#     'globaltoc_includehidden': "true",
#
#     # HTML navbar class (Default: "navbar") to attach to <div> element.
#     # For black navbar, do "navbar navbar-inverse"
#     'navbar_class': "navbar navbar-inverse",
#
#     # Fix navigation bar to top of page?
#     # Values: "true" (default) or "false"
#     'navbar_fixed_top': "true",
#
#     # Location of link to source.
#     # Options are "nav" (default), "footer" or anything else to exclude.
#     'source_link_position': "nav",
#
#     # Bootswatch (http://bootswatch.com/) theme.
#     #
#     # Options are nothing (default) or the name of a valid theme
#     # such as "cosmo" or "sandstone".
#     #
#     # The set of valid themes depend on the version of Bootstrap
#     # that's used (the next config option).
#     #
#     # Currently, the supported themes are:
#     # - Bootstrap 2: https://bootswatch.com/2
#     # - Bootstrap 3: https://bootswatch.com/3
#     'bootswatch_theme': "cosmo",
#
#     # Choose Bootstrap version.
#     # Values: "3" (default) or "2" (in quotes)
#     'bootstrap_version': "3",
# }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'melusinedoc'


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'melusine.tex',
     u'melusine Documentation',
     u'Sacha Samama & Tom Stringer', 'manual'),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'melusine',
     u'Melusine Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'melusine',
     u'Melusine Documentation',
     author,
     'Melusine',
     'One line description of project.',
     'Miscellaneous'),
]



import keras
from keras.layers import BatchNormalization
from keras.layers import LeakyReLU
from keras.layers import Input
from keras.layers import concatenate
from keras.optimizers import Adam
from keras.models import Model
from keras.layers import Dense
from keras.layers import Conv1D
from keras.layers import Dropout
from keras.layers import SpatialDropout1D
from keras.layers import GlobalMaxPooling1D
from keras.layers import GRU
from keras.layers import Bidirectional


def cnn_model(embedding_matrix_init,
              ntargets=18,
              seq_max=100,
              nb_meta=252,
              loss='categorical_crossentropy'):
    """Pre-defined architecture of a CNN model.

    Parameters
    ----------
    embedding_matrix_init : np.array,
        Pretrained embedding matrix.

    ntargets : int, optional
        Dimension of model output.
        Default value, 18.

    seq_max : int, optional
        Maximum input length.
        Default value, 100.

    nb_meta : int, optional
        Dimension of meta data input.
        Default value, 252.

    loss : str, optional
        Loss function for training.
        Default value, 'categorical_crossentropy'.

    Returns
    -------
    Model instance
    """

    text_input = Input(shape=(seq_max,), dtype='int32')

    x = keras.layers.Embedding(input_dim=embedding_matrix_init.shape[0],
                               output_dim=embedding_matrix_init.shape[1],
                               input_length=seq_max,
                               weights=[embedding_matrix_init],
                               trainable=True)(text_input)
    x = Conv1D(200, 2, padding='same', activation='linear', strides=1)(x)
    x = SpatialDropout1D(0.15)(x)
    x = BatchNormalization()(x)
    x = LeakyReLU(alpha=0.05)(x)
    x = Conv1D(250, 2, padding='same', activation='linear', strides=1)(x)
    x = SpatialDropout1D(0.15)(x)
    x = LeakyReLU(alpha=0.05)(x)
    x = Dropout(0.15)(x)
    x = GlobalMaxPooling1D()(x)
    x = Dense(250, activation="linear")(x)
    x = LeakyReLU(alpha=0.05)(x)
    x = Dense(150, activation="linear")(x)
    x = Dropout(0.15)(x)
    x = LeakyReLU(alpha=0.05)(x)

    if nb_meta == 0:
        inputs = text_input
        concatenate_2 = x
    else:
        Meta_input = Input(shape=(nb_meta,), dtype='float32')
        inputs = [text_input, Meta_input]

        concatenate_1 = Meta_input
        y = Dense(150, activation="linear")(concatenate_1)
        y = Dropout(0.2)(y)
        y = LeakyReLU(alpha=0.05)(y)
        y = Dense(100, activation="linear")(y)
        y = Dropout(0.2)(y)
        y = LeakyReLU(alpha=0.05)(y)
        y = Dense(80, activation="linear")(y)
        y = Dropout(0.2)(y)
        y = LeakyReLU(alpha=0.05)(y)
        concatenate_2 = concatenate([x, y])

    z = Dense(200, activation="linear")(concatenate_2)
    z = Dropout(0.2)(z)
    z = LeakyReLU(alpha=0.05)(z)
    z = Dense(100, activation="linear")(z)
    z = Dropout(0.2)(z)
    z = LeakyReLU(alpha=0.05)(z)
    outputs = Dense(ntargets, activation='softmax')(z)

    model = Model(inputs=inputs,
                  outputs=outputs)

    model.compile(optimizer=Adam(),
                  loss=loss,
                  metrics=['accuracy'])

    return model


def rnn_model(embedding_matrix_init,
              ntargets=18,
              seq_max=100,
              nb_meta=252,
              loss='categorical_crossentropy'):
    """Pre-defined architecture of a RNN model.

    Parameters
    ----------
    embedding_matrix_init : np.array,
        Pretrained embedding matrix.

    ntargets : int, optional
        Dimension of model output.
        Default value, 18.

    seq_max : int, optional
        Maximum input length.
        Default value, 100.

    nb_meta : int, optional
        Dimension of meta data input.
        Default value, 252.

    loss : str, optional
        Loss function for training.
        Default value, 'categorical_crossentropy'.

    Returns
    -------
    Model instance
    """
    text_input = Input(shape=(seq_max,), dtype='int32')

    x = keras.layers.Embedding(input_dim=embedding_matrix_init.shape[0],
                               output_dim=embedding_matrix_init.shape[1],
                               input_length=seq_max,
                               weights=[embedding_matrix_init],
                               trainable=True)(text_input)
    x = Bidirectional(GRU(80, return_sequences=True))(x)
    x = SpatialDropout1D(0.15)(x)
    x = Bidirectional(GRU(40, return_sequences=True))(x)
    x = SpatialDropout1D(0.15)(x)
    x = GlobalMaxPooling1D()(x)
    x = Dense(250, activation="linear")(x)
    x = LeakyReLU(alpha=0.05)(x)
    x = Dense(150, activation="linear")(x)
    x = Dropout(0.15)(x)
    x = LeakyReLU(alpha=0.05)(x)

    if nb_meta == 0:
        inputs = text_input
        concatenate_2 = x
    else:
        Meta_input = Input(shape=(nb_meta,), dtype='float32')
        inputs = [text_input, Meta_input]

        concatenate_1 = Meta_input
        y = Dense(150, activation="linear")(concatenate_1)
        y = Dropout(0.2)(y)
        y = LeakyReLU(alpha=0.05)(y)
        y = Dense(100, activation="linear")(y)
        y = Dropout(0.2)(y)
        y = LeakyReLU(alpha=0.05)(y)
        y = Dense(80, activation="linear")(y)
        y = Dropout(0.2)(y)
        y = LeakyReLU(alpha=0.05)(y)
        concatenate_2 = concatenate([x, y])

    z = Dense(200, activation="linear")(concatenate_2)
    z = Dropout(0.2)(z)
    z = LeakyReLU(alpha=0.05)(z)
    z = Dense(100, activation="linear")(z)
    z = Dropout(0.2)(z)
    z = LeakyReLU(alpha=0.05)(z)
    output = Dense(ntargets, activation='softmax')(z)

    model = Model(inputs=inputs,
                  outputs=output)

    model.compile(optimizer=Adam(),
                  loss=loss,
                  metrics=['accuracy'])

    return model



import pandas as pd
from melusine.prepare_email.build_historic import build_historic

body = " \n  \n  \n Bonjours, \n  \n Suite a notre conversation téléphonique \
de Mardi , pourriez vous me dire la \n somme que je vous dois afin d'd'être \
en régularisation . \n  \n Merci bonne journée \n  \n Le mar. 22 mai 2018 \
à 10:20,  <gestionsocietaire@mutuelle.fr> a écrit\xa0: \n Bonjour. \n  \n \
Merci de bien vouloir prendre connaissance du document ci-joint : \n 1 - \
Relevé d'identité postal MUTUELLE (contrats) \n  \n Sentiments \
mutualistes. \n  \n La Mututelle \n  \n La visualisation des fichiers PDF \
nécessite Adobe Reader. \n  "


output = [
    {'text': " \n  \n  \n Bonjours, \n  \n Suite a notre conversation \
téléphonique de Mardi , pourriez vous me dire la \n somme que je vous \
dois afin d'd'être en régularisation . \n  \n Merci bonne journée",
     'meta': ''},
    {'text': " \n Bonjour. \n  \n Merci de bien vouloir prendre connaissance \
du document ci-joint : \n 1 - Relevé d'identité postal MUTUELLE \
(contrats) \n  \n Sentiments mutualistes. \n  \n La Mututelle \n  \n \
La visualisation des fichiers PDF nécessite Adobe Reader. \n  ",
     'meta': ' \n  \n Le mar. 22 mai 2018 à 10:20,  \
<gestionsocietaire@mutuelle.fr> a écrit\xa0:'}]


def test_build_historic():
    input_df = pd.DataFrame({
        'body': [body]
    })

    output_df = pd.Series([output])

    result = input_df.apply(build_historic, axis=1)
    pd.testing.assert_series_equal(result, output_df)



import json
import os.path as op
import logging
from configparser import ConfigParser
import pandas as pd
import unidecode

class ConfigJsonReader():
    """Class to initialize a config.json file that contains your own
    parameters used by the package.

    In fact, the module `prepare_email` contains a set of functions which use "parameters"
    pre-defined in the conf.json.

    The json file contains parameters, such as:
        - general or specific regular expressions
        - footers, greetings or other mail keywords/tags

    Attributes
    ----------
    path_ini_file_ : str,
        Path to the path.ini file (file used only to check,
        set or get path of conf.json).

    path_default_conf_json_ : str,
        Path to the default conf.json (define by the contributors).

    Examples
    --------
    >>> conf = ConfigJsonReader()
    >>> conf.path_ini_file_   # will return path to the created .ini file

    >>> # I defined my own path otherwise it uses a default conf.json
    >>> conf.set_config_path(file_path='path/to/my/conf.json')
    >>> conf_dict = conf.get_config_file()
    >>> print(conf_dict)  # will print the json.

    """

    def __init__(self):
        config_directory = op.dirname(op.abspath(__file__))
        self.path_ini_file_ = op.join(config_directory, 'path.ini')
        self.path_default_conf_json_ = op.join(config_directory, 'conf.json')
        self.path_default_names_csv_ = op.join(config_directory, 'names.csv')

        if not op.exists(self.path_ini_file_):
            logging.info("Create an path.ini file to configure your own config.json")
            ini_file = open(self.path_ini_file_, 'w')
            conf = ConfigParser()
            conf.add_section('PATH')
            conf.set('PATH', 'template_config', self.path_default_conf_json_)
            conf.set('PATH', 'default_name_file', self.path_default_names_csv_)
            conf.write(ini_file)
            ini_file.close()

        self.config = ConfigParser()
        self.config.read(self.path_ini_file_)

        pass

    def set_config_path(self, file_path=None):
        """Set a path for your own `config.json`.

        Parameters
        ----------
        file_path: str, optional
            Path to the json file. If set to None (default), it will use the default
            json located in the built-in package `melusine/config/conf.json`.
        """
        if file_path is not None:
            # if file_path is specified, it writes new path in the .ini file.
            self.config['PATH']['template_config'] = file_path
            with open(self.path_ini_file_, 'w') as configfile:
                self.config.write(configfile)
        pass

    def get_config_file(self):
        """Load a config json file from the given path.
           Load the list of names from the names.csv file.
        """
        path = self.config['PATH']['template_config']
        if path == self.path_default_conf_json_:
            config_file = self.load_config_file(path=None)
        else:
            config_file = self.load_config_file(path=path)
        name_file_path = self.config['PATH']['default_name_file']

        if name_file_path == self.path_default_names_csv_:
            name_list = self.load_name_file(path=None)
        else:
            name_list = self.load_name_file(path=name_file_path)

        config_file['words_list']['names'] = name_list

        return config_file

    def reset_config_path(self):
        self.config['PATH']['template_config'] = self.path_default_conf_json_
        with open(self.path_ini_file_, 'w') as configfile:
            self.config.write(configfile)

        pass

    def load_config_file(self, path=None):
        """Load Json."""
        # by default it takes native the config.json
        if path is None:
            path = self.path_default_conf_json_

        with open(file=path, mode='r') as file:
            config_file = json.load(file)

        return config_file

        logging.info("Load config from path: {}.".format(path))

    def set_name_file_path(self, file_path=None):
        """Set a path for your own `names.csv`.

        Parameters
        ----------
        file_path: str, optional
            Path to the csv file. If set to None (default), it will use the default
            csv located in the built-in package `melusine/config/names.csv`.
        """
        if file_path is not None:
            # if file_path is specified, it writes new path in the .ini file.
            self.config['PATH']['default_name_file'] = file_path
            with open(self.path_ini_file_, 'w') as configfile:
                self.config.write(configfile)
        pass

    def reset_name_file_path(self):
        self.config['PATH']['default_name_file'] = self.path_default_names_csv_
        with open(self.path_ini_file_, 'w') as configfile:
            self.config.write(configfile)
        pass

    def load_name_file(self, path):
        """Load csv."""
        # by default it takes native the names.csv
        if path is None:
            path = self.path_default_names_csv_

        try:
            df_names = pd.read_csv(path, encoding="latin-1", sep=";")
            name_list = df_names['Name'].values
            name_list = [unidecode.unidecode(p).lower() for p in name_list]
        except FileNotFoundError:
            name_list = []

        return name_list



import os.path as op
import pandas as pd


def load_email_data():
    """
    Load a csv file containing toy email data

    Return
    ------
    pandas.DataFrame
        DataFrame with toy email data
    """

    data_directory = op.dirname(op.abspath(__file__))
    email_data_path = op.join(data_directory, 'emails.csv')

    return pd.read_csv(email_data_path, encoding='utf-8', sep=';')




def print_color_mail(structured_body):
    """Highlight the tagged sentences.

    Parameters
    ----------
    structured_body : a structured body from process_sent_tag,

    Returns
    -------
    Print the mail by sentence.

    """
    for message in structured_body:
        print("___________________________\n")
        print_color(str(message.get("meta")), "META")

        structured_text = message.get("structured_text")
        header = structured_text.get("header")
        print_color(str(header), "HEADER")

        for sentence in structured_text.get("text"):
            text = sentence.get("part")
            tag = sentence.get("tags")
            print_color(text, tag)


def print_color(text, part=None):
    """Select according to the tag the right color to use when printing."""
    switcher_tag = {
        "HELLO": "\033[0;37;44m" + "HELLO" + "\033[0m",
        "GREETINGS": "\033[0;37;45m" + "GREETINGS" + "\033[0m",
        "SIGN": "\033[0;37;41m" + "SIGN" + "\033[0m",
        "THANKS": "\033[0;31;46m" + "THANKS" + "\033[0m",
        "PJ": "\033[0;37;42m" + "PJ" + "\033[0m",
        "META": "\33[43m" + "META" + "\033[0m",
        "FOOTER": "\33[41m" + "FOOTER" + "\033[0m",
        "DISCLAIMER": "\33[41m" + "DISCLAIMER" + "\033[0m",
        "TYPO": "\33[47m" + "TYPO" + "\033[0m",
        "HEADER": "\033[0;37;41m" + "HEADER" + "\033[0m"
    }

    switcher = {
        "HELLO": "\033[0;37;44m" + text + "\033[0m",
        "GREETINGS": "\033[0;37;45m" + text + "\033[0m",
        "SIGN": "\033[0;37;41m" + text + "\033[0m",
        "THANKS": "\033[0;31;46m" + text + "\033[0m",
        "PJ": "\033[0;37;42m" + text + "\033[0m",
        "META": "\33[43m" + text + "\033[0m",
        'FOOTER': "\33[41m" + text + "\033[0m",
        'DISCLAIMER': "\33[41m" + text + "\033[0m",
        'TYPO': "\33[47m" + text + "\033[0m",
        "HEADER": "\033[0;37;41m" + text + "\033[0m"
    }

    # print("TAG : ", switcher_tag.get(part, text))
    if part == "BODY":
        print("> BODY : ", switcher.get(part, text))
    else:
        print("> ", switcher_tag.get(part, text), " : ", switcher.get(part, text))



from melusine.prepare_email.mail_segmenting import split_message_to_sentences
from melusine.utils.multiprocessing import apply_by_multiprocessing
from melusine.nlp_tools.tokenizer import Tokenizer
from melusine.config.config import ConfigJsonReader

conf_reader = ConfigJsonReader()

class Streamer():
    """Class to transform pd.Series into stream.

    Used to prepare the data for the training of the phraser and embeddings.

    Attributes
    ----------
    column : str,
        Input text column(s) to consider for the streamer.

    stream : MailIterator object,
        Stream of all the tokens of the pd.Series.

    Examples
    --------
    >>> streamer = Streamer()
    >>> streamer.to_stream(X) # will build the stream attribute
    >>> tokens_stream =  = streamer.stream
    >>> print(tokens_stream)

    """

    def __init__(self, stop_removal=False, column='clean_body', n_jobs=1):
        self.column_ = column
        self.n_jobs = n_jobs
        config = conf_reader.get_config_file()
        stopwords = config["words_list"]["stopwords"] + config["words_list"]["names"]
        self.tokenizer = Tokenizer(stopwords, stop_removal=stop_removal)

    def to_stream(self, X):
        """Build a MailIterator object containing a stream of tokens from
        a pd.Series.

        Parameters
        ----------
        X : pd.Dataframe.

        Examples
        --------
        >>> streamer.to_stream(X) # will build the stream attribute
        >>> tokens_stream =  = streamer.stream
        >>> print(tokens_stream)

        """
        flattoks = self.to_flattoks(X)
        self.stream = MailIterator(flattoks)
        pass

    def to_flattoks(self, X):
        """Create list of list of tokens from a pd.Series
        Each list of tokens correspond to a sentence.

        Parameters
        ----------
        X : pd.Dataframe,

        Returns
        -------
        list of lists of strings
        """
        tokenized_sentences_list = apply_by_multiprocessing(df=X[[self.column_]],
                                                            func=lambda x: self.to_list_of_tokenized_sentences(x[self.column_]),
                                                            args=None,
                                                            workers=self.n_jobs,
                                                            progress_bar=False
                                                            )
        flattoks = [item for sublist in tokenized_sentences_list
                    for item in sublist]
        return flattoks

    def to_list_of_tokenized_sentences(self, text):
        """Create list of list of tokens from a text.
        Each list of tokens correspond to a sentence.

        Parameters
        ----------
        text : str

        Returns
        -------
        list of list of strings
        """
        #text = row[self.column_]
        sentences_list = split_message_to_sentences(text)
        tokenized_sentences_list = [self.tokenizer._tokenize(sentence)
                                    for sentence in sentences_list
                                    if sentence != ""]
        return tokenized_sentences_list


class MailIterator():
    """Class to transform stream of tokens into iterators."""

    def __init__(self, tok_stream):
        self.tok_stream = tok_stream

    def __iter__(self):
        for sent in self.tok_stream:
            yield sent
